#include "Curve.h"

#include <algorithm>

#include <functional>

#include <assert.h>

#include <math.h>

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>

#include "Curves.h"

#include "CurveEvaluator.h"

#include "CurveEvaluators.h"

#include "UniqueColorGenerator.h"

#include "scene.h"

#include "aniUI.h"

#include "fltkCurveEditor.h"

extern CAniUI *aniUI;

extern UniqueColorGenerator *gUniqueColorGenerator;
extern CurveEvaluators *gCurveEvaluators;

extern bool g_viewCurve;
extern bool g_viewBezier;

Curve::Curve(void)
		:m_name("notnamed"),
		 m_current_curve_type("Linear"),
		 m_currently_shown(false),
		 m_wrap_control_points(false),
		 m_dirty(true)
{
	performNewCurveSetup();
}

Curve::Curve(const std::string & name, const double & start_value)
		:m_name(name),
		 m_current_curve_type("Linear"),
		 m_currently_shown(false),
		 m_wrap_control_points(false),
		 m_dirty(true)
{
	performNewCurveSetup(start_value);
}

void Curve::performNewCurveSetup(const double & start_value)
{
	m_color = UniqueColorGenerator::getInstance()->getUniqueColor();

	m_control_points.push_back(Point(8.0, start_value));
	m_control_points.push_back(Point(12.0, start_value));

	assert(m_control_points.size() == 2);
}

Curve::Curve(std::istream & input_stream)
		:m_dirty(true)
{
	fromStream(input_stream);
}

void Curve::toStream(std::ostream & output_stream) const
{
	output_stream << m_name << std::endl;
	
	output_stream << m_control_points.size() << std::endl;

	for (std::vector<Point>::const_iterator control_point_iterator = m_control_points.begin(); control_point_iterator != m_control_points.end(); ++control_point_iterator)
	{
		output_stream << *control_point_iterator;
	}

	output_stream << m_wrap_control_points << std::endl;

	output_stream << m_current_curve_type << std::endl;

	output_stream << m_color.red << std::endl;
	output_stream << m_color.green << std::endl;
	output_stream << m_color.blue << std::endl;

	output_stream << m_zoom.getLeft() << std::endl;
	output_stream << m_zoom.getRight() << std::endl;

	output_stream << m_zoom.getBottom() << std::endl;
	output_stream << m_zoom.getTop() << std::endl;
}

void Curve::fromStream(std::istream & input_stream)
{
	std::getline(input_stream, m_name);
	std::getline(input_stream, m_name);
	
	unsigned int number_of_control_points;

	input_stream >> number_of_control_points;

	m_control_points.resize(number_of_control_points);

	for (unsigned int control_point_index = 0; control_point_index < number_of_control_points; ++control_point_index)
	{
		input_stream >> m_control_points[control_point_index];
	}

	input_stream >> m_wrap_control_points;
	aniUI->m_pCurveEditor->setCurveButton(m_wrap_control_points);
	
	std::getline(input_stream, m_current_curve_type);
	std::getline(input_stream, m_current_curve_type);

	input_stream >> m_color.red;
	input_stream >> m_color.green;
	input_stream >> m_color.blue;

	double left_zoom;

	input_stream >> left_zoom;

	double right_zoom;

	input_stream >> right_zoom;

	double bottom_zoom;

	input_stream >> bottom_zoom;

	double top_zoom;

	input_stream >> top_zoom;

	m_zoom.setupObjectSpaceDimensions(left_zoom, right_zoom, bottom_zoom, top_zoom);

	m_dirty = true;
}

std::string Curve::getName(void) const
{
	return m_name;
}

void Curve::setCurveType(const std::string & curve_type)
{
	m_current_curve_type = curve_type;
	if(m_name == viewpoint)
	{
		Curves::getInstance()->getCurve(qnowx)->setCurveType(curve_type);
		Curves::getInstance()->getCurve(qnowy)->setCurveType(curve_type);
		Curves::getInstance()->getCurve(qnowz)->setCurveType(curve_type);
		Curves::getInstance()->getCurve(qnoww)->setCurveType(curve_type);
		Curves::getInstance()->getCurve(centerx)->setCurveType(curve_type);
		Curves::getInstance()->getCurve(centery)->setCurveType(curve_type);
		Curves::getInstance()->getCurve(centerz)->setCurveType(curve_type);
		Curves::getInstance()->getCurve(zoomed)->setCurveType(curve_type);
	}

	m_dirty = true;
}

void Curve::setWrapControlPointsFlag(const bool & wrap_control_points)
{
	m_wrap_control_points = wrap_control_points;

	m_dirty = true;
}

double Curve::evaluateCurveAt(const double & x) const
{
	fixIfDirty();
	
	double value = 0.0;

	if (m_evaluated_curve_points.size() > 1)
	{
		std::vector<Point>::iterator first_point = m_evaluated_curve_points.begin();
		std::vector<Point>::iterator  last_point = m_evaluated_curve_points.end() - 1;

		bool evaluate_point_to_left_of_range	= (first_point->x	> x);
		bool evaluate_point_to_right_of_range	= (last_point->x	< x);

		if (evaluate_point_to_left_of_range)
		{
			value = first_point->y;
		}
		else if (evaluate_point_to_right_of_range)
		{
			value = last_point->y;
		}
		else
		{
			std::vector<Point>::iterator point_one_iterator = first_point;

			while ((point_one_iterator + 1)->x < x)
			{
				++point_one_iterator;

				assert(point_one_iterator != last_point);
			}

			std::vector<Point>::iterator point_two_iterator = point_one_iterator + 1;
			
			assert(point_one_iterator != m_evaluated_curve_points.end());
			assert(point_two_iterator != m_evaluated_curve_points.end());

			Point point_one = *point_one_iterator;
			Point point_two = *point_two_iterator;
			
			double slope = (point_two.y - point_one.y) / (point_two.x - point_one.x);

			value = (x - point_one.x) * slope + point_one.y;
		}
	}

	return value;
}

double Curve::getMinimumX(void) const
{
	fixIfDirty();

	std::vector<Point>::const_iterator minimum = std::min_element(m_evaluated_curve_points.begin(), m_evaluated_curve_points.end(), PointSmallerXCompare());

	return minimum->x;
}

double Curve::getMaximumX(void) const
{
	fixIfDirty();

	std::vector<Point>::const_iterator maximum = std::max_element(m_evaluated_curve_points.begin(), m_evaluated_curve_points.end(), PointLargerXCompare());

	return maximum->x;
}

void Curve::scaleX(const double & scaling_factor)
{
	for (std::vector<Point>::iterator control_point_iterator = m_control_points.begin(); control_point_iterator != m_control_points.end(); ++control_point_iterator)
	{
		control_point_iterator->x *= scaling_factor;
	}

	m_dirty = true;
}

void Curve::addControlPoint(const Point & normalized_point)
{
	Point point(m_zoom.getObjectSpaceX(normalized_point.x), m_zoom.getObjectSpaceY(normalized_point.y));

	unsigned int old_number_of_control_points = m_control_points.size();
	
	m_control_points.push_back(point);

	unsigned int new_number_of_control_points = m_control_points.size();

	assert(new_number_of_control_points == (old_number_of_control_points + 1));

	m_dirty = true;
}

void Curve::forcePoint(double x, double y)
{
	Point point(x,y);
	m_control_points.push_back(point);
	m_dirty = true;
}

void Curve::removeControlPoint(const unsigned int & control_point_index)
{
	assert (0 <= control_point_index);
	assert(control_point_index < m_control_points.size());
	
	std::vector<Point>::const_iterator erase_result;
	
	erase_result = m_control_points.erase(m_control_points.begin() + control_point_index);
	if(m_name == viewpoint)
	{
		Curves::getInstance()->getCurve(qnowx)->removeControlPoint(control_point_index);
		Curves::getInstance()->getCurve(qnowy)->removeControlPoint(control_point_index);
		Curves::getInstance()->getCurve(qnowz)->removeControlPoint(control_point_index);
		Curves::getInstance()->getCurve(qnoww)->removeControlPoint(control_point_index);
		Curves::getInstance()->getCurve(centerx)->removeControlPoint(control_point_index);
		Curves::getInstance()->getCurve(centery)->removeControlPoint(control_point_index);
		Curves::getInstance()->getCurve(centerz)->removeControlPoint(control_point_index);
		Curves::getInstance()->getCurve(zoomed)->removeControlPoint(control_point_index);
	}
	m_dirty = true;

}

unsigned int Curve::getClosestControlPoint(const Point & normalized_point, double * closest_distance) const
{
	fixIfDirty();
	
	Point point(m_zoom.getObjectSpaceX(normalized_point.x), m_zoom.getObjectSpaceY(normalized_point.y));

	Point * closest_control_point = NULL;

	static const double MAXIMUM_DISTANCE = 9e99;

	double closest_distance_squared = MAXIMUM_DISTANCE;

	if (m_currently_shown)
	{
		for (std::vector<Point>::iterator control_point_iterator = m_control_points.begin(); control_point_iterator != m_control_points.end(); ++control_point_iterator)
		{
			double delta_x = (control_point_iterator->x - point.x);
			double delta_y = (control_point_iterator->y - point.y);

			double distance_squared = delta_x * delta_x + delta_y * delta_y;

			if (distance_squared < closest_distance_squared)
			{
				closest_control_point		= control_point_iterator;
				closest_distance_squared	= distance_squared;
			}
		}
	}

	double delta_x = normalized_point.x - ((closest_control_point->x - m_zoom.getLeft()		) / (m_zoom.getRight()	- m_zoom.getLeft())		);
	double delta_y = normalized_point.y - ((closest_control_point->y - m_zoom.getBottom()	) / (m_zoom.getTop()	- m_zoom.getBottom())	);

	*closest_distance = sqrt(delta_x * delta_x + delta_y * delta_y);

	return (closest_control_point - m_control_points.begin());
}

double Curve::getDistanceToCurve(const Point & normalized_point) const
{
	fixIfDirty();

	Point point(m_zoom.getObjectSpaceX(normalized_point.x), m_zoom.getObjectSpaceY(normalized_point.y));

	Point evaluated_point(point.x, evaluateCurveAt(point.x));

	double delta_x = normalized_point.x - ((evaluated_point.x  - m_zoom.getLeft()	) / (m_zoom.getRight()	- m_zoom.getLeft())	);
	double delta_y = normalized_point.y - ((evaluated_point.y  - m_zoom.getBottom()	) / (m_zoom.getTop()	- m_zoom.getBottom())	);

	double distance = sqrt(delta_x * delta_x + delta_y * delta_y);

	return distance;
}

Point Curve::getControlPoint(const unsigned int & control_point_index) const
{
	assert (0 <= control_point_index);
	assert(control_point_index < m_control_points.size());

	assert(m_currently_shown);

	return m_control_points[control_point_index];
}

unsigned int Curve::getNumberOfControlPoints(void) const
{
	return m_control_points.size();
}

unsigned int Curve::getNumberOfEvaluatedCurveSegments(void) const
{
	fixIfDirty();

	unsigned int number_of_evaluated_control_points = m_evaluated_curve_points.size();

	unsigned int number_of_curve_segments;

	if (number_of_evaluated_control_points == 0)
	{
		number_of_curve_segments = 0;
	}
	else
	{
		number_of_curve_segments = number_of_evaluated_control_points - 1;
	}

	return number_of_curve_segments;
}

bool Curve::isShown(void) const
{
	return m_currently_shown;
}

void Curve::changeControlPoint(const unsigned int & control_point_index, const Point & normalized_new_point)
{
	unsigned int number_of_control_points = m_control_points.size();

	assert (0 <= control_point_index);
	assert(control_point_index < number_of_control_points);

	assert(m_currently_shown);

	Point new_point(m_zoom.getObjectSpaceX(normalized_new_point.x), m_zoom.getObjectSpaceY(normalized_new_point.y));
	
	m_control_points[control_point_index] = new_point;

	m_dirty = true;
}

bool Curve::toggleShowCurve(void)
{
	m_currently_shown = ! m_currently_shown;

	m_dirty = true;

	return m_currently_shown;
}

void Curve::drawCurve(void) const
{
	if (m_currently_shown)
	{
		fixIfDirty();

		m_zoom.applyTopZoom();

		glColor3d(m_color.red, m_color.green, m_color.blue);

		drawEvaluatedCurveSegments();
	}
}

void Curve::drawEvaluatedCurveSegments(void) const
{
	if (m_currently_shown)
	{
		fixIfDirty();

		m_zoom.applyTopZoom();
		glBegin(GL_LINE_STRIP);

		std::vector<Point>::const_iterator evaluated_curve_point_iterator = m_evaluated_curve_points.begin(); 
		for (;
			     evaluated_curve_point_iterator != m_evaluated_curve_points.end(); 
			     ++evaluated_curve_point_iterator)
			{
				glVertex2d(evaluated_curve_point_iterator->x, evaluated_curve_point_iterator->y);

			}
		
		glEnd();
		if(g_viewCurve)
		{
			glPointSize(3);
			glColor3f(0,0,0);
			glBegin(GL_POINTS);

				for ( evaluated_curve_point_iterator = m_evaluated_curve_points.begin(); evaluated_curve_point_iterator != m_evaluated_curve_points.end(); ++evaluated_curve_point_iterator)
				{
					glVertex2d(evaluated_curve_point_iterator->x, evaluated_curve_point_iterator->y);

				}
			glEnd();
			glColor3f(1,1,1);

		}
		if(g_viewBezier)
		{
			glPointSize(5);
			glColor3f(1,0,0);
			glBegin(GL_POINTS);

				for ( std::vector<Point>::const_iterator bezier_curve_point_iterator = m_bezier_curve_points.begin(); bezier_curve_point_iterator != m_bezier_curve_points.end(); ++bezier_curve_point_iterator)
				{
					glVertex2d(bezier_curve_point_iterator->x, bezier_curve_point_iterator->y);

				}
			glEnd();
			glColor3f(1,1,1);
		}
	}
}

void Curve::drawControlPoints(void) const
{
	if (m_currently_shown)
	{
		fixIfDirty();

		m_zoom.applyTopZoom();

		static const unsigned int CONTROL_POINT_SIZE = 5;

		glPointSize(CONTROL_POINT_SIZE);

		glBegin(GL_POINTS);

			for (std::vector<Point>::const_iterator control_point_iterator = m_control_points.begin(); control_point_iterator != m_control_points.end(); ++control_point_iterator)
			{
				glVertex2d(control_point_iterator->x, control_point_iterator->y);
			}

		glEnd();
		
		glPointSize(1);
	}
}

Zoomer * Curve::getZoomer(void)
{
	return &m_zoom;
}

const Zoomer * Curve::getZoomer(void) const
{
	return &m_zoom;
}

void Curve::fixIfDirty(void) const
{
	if (m_dirty)
	{
#ifdef WIN32
		m_evaluated_curve_points.assign(m_evaluated_curve_points.size());
		m_bezier_curve_points.assign(m_bezier_curve_points.size());

#else
		int n=m_evaluated_curve_points.size(), i;
		for (i=0; i<n; i++)
			m_evaluated_curve_points.pop_back();
		for (i=0; i<n; i++)
			m_evaluated_curve_points.push_back(Point());
		n=m_bezier_curve_points.size(), i;
		for (i=0; i<n; i++)
			m_bezier_curve_points.pop_back();
		for (i=0; i<n; i++)
			m_bezier_curve_points.push_back(Point());

#endif
		unsigned int number_of_control_points = m_control_points.size();

		std::vector<Point> sorted_control_points = m_control_points;

		std::sort(	sorted_control_points.begin(),
					sorted_control_points.end(),
					PointSmallerXCompare());

		unsigned int number_of_sorted_control_points = sorted_control_points.size();

		CurveEvaluators * curve_evaluators = CurveEvaluators::getInstance();
		
		CurveEvaluator * curve_evaluator = curve_evaluators->getCurveEvaluator(m_current_curve_type);

		curve_evaluator->evaluateCurve(sorted_control_points, m_evaluated_curve_points, m_zoom.getRight(), m_wrap_control_points, m_bezier_curve_points);

		std::sort(	m_evaluated_curve_points.begin(),
					m_evaluated_curve_points.end(),
					PointSmallerXCompare());
		std::sort(	m_bezier_curve_points.begin(),
					m_bezier_curve_points.end(),
					PointSmallerXCompare());


		m_dirty = false;
	}
}

void Curve::markDirty(void) const
{
	m_dirty = true;
}

std::ostream & operator<<(std::ostream & output_stream, const Curve & curve_data)
{
	curve_data.toStream(output_stream);

	return output_stream;
}

std::istream & operator>>(std::istream & input_stream, Curve & curve_data)
{
	curve_data.fromStream(input_stream);

	return input_stream;
}

